"""
    Billets - a module for OpenERP to help organise billets for events.
    Copyright (C) 2011  Andrew Miller <andrew@amxl.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import tools
from osv import fields,osv,orm

flexibility_options = [
    ('flex', 'Either'),
    ('defno', 'No'),
    ('defyes', 'Yes'),
    ('flexno', 'Flexible - prefer not'),
    ('flexyes', 'Flexible - prefer yes')]

match_status = [
      ('none', 'Unmatched'),
      ('provaut', 'Provisional match (automated)'),
      ('provman', 'Provisional match (manual)'),
      ('final', 'Final match')
    ]

notify_status = [
      ('none', 'Not yet notified of match'),
      ('email', 'E-mail sent about match'),
      ('confirm', 'Suitability confirmed'),
      ('refuse', 'Match rejected')
    ]


def brute_force_match(s, v):
    for (a, o, b) in s:
        val = v[a]
        if o == '=':
            return val == b
        elif o == '!=' or o == '<>':
            return val != b
        elif o == '<':
            return val < b
        elif o == '<=':
            return val <= b
        elif o == '>':
            return val > b
        elif o == '>=':
            return val >= b
        else:
            raise 'Unsupported operator on brute force search'

def brute_force_search(self, cr, uid, obj, name, args, context=None):
    s = [(a,o,b) for (a,o,b) in args if a==name]
    rem = [(a,o,b) for (a,o,b) in args if a != name]
    ids = self.search(cr, uid, rem)
    ret = [v.id for v in self.browse(cr, uid, ids) if brute_force_match(s, v)]
    #print "brute_force_search(%s, %s, %s, %s, %s, %s, %s) = %s" %\
    #    (self, cr, uid, obj, name, args, context, ret)
    return [('id','in',ret)]

class billet_event(osv.osv):
    _name = 'billets.event'
    _description = 'Event that requires billets'
    _columns = {
          'name': fields.char('Event Name', size=20)
        }
    _defaults = {}
billet_event()

class billet_day(osv.osv):
    _name = 'billets.day'
    _description = 'Day for billet hosting'
    _columns = {
          'name': fields.char('Day Name', size=20),
          'event': fields.many2one('billets.event', 'Event', required=True)
        }
    _defaults = {}
billet_day()

class billet_case(osv.osv):
    _name = 'billets.case'
    _description = 'Billet Case'

    def _get_case_name(self, cr, uid, ids, field_name, arg, context):
        ret = {}
        for c in self.browse(cr, uid, ids):
            ret[c.id] = c.partner_id.name
        return ret

    _columns = {
          'name': fields.function(_get_case_name, fnct_search=brute_force_search, method=True, string='Name', type='text'),
          'event': fields.many2one('billets.event', 'Event', required=True),
          'partner_id': fields.many2one('res.partner', 'Person', required=True),
          'lacto_ovo_veg': fields.selection(flexibility_options, 'Lacto-ovo vegetarian', size=16, required=True),
          'strict_vegan': fields.selection(flexibility_options, 'Strict vegan', size=16, required=True),
          'gluten_free': fields.selection(flexibility_options, 'Gluten free', size=16, required=True),
          'omnivore': fields.selection(flexibility_options, 'Omnivore', size=16, required=True),
          'low_carb': fields.selection(flexibility_options, 'Low carbohydrate', size=16, required=True),
          'other_diets': fields.selection(flexibility_options, 'Other dietary requirements', size=16, required=True),
          'transport': fields.selection(flexibility_options, 'Provides transport', size=16, required=True),
          'details': fields.text('Details')
        }
    _defaults = { 'lacto_ovo_veg': lambda *a: 'flex',
                  'strict_vegan': lambda *a: 'flex',
                  'gluten_free': lambda *a: 'flex',
                  'omnivore': lambda *a: 'flex',
                  'low_carb': lambda *a: 'flex',
                  'other_diets': lambda *a: 'flex',
                  'transport': lambda *a: 'flex'
                }
billet_case()

class billet_accommodation_offered(osv.osv):
    _name = 'billets.offered'
    _description = 'Billet Accommodation Offered'
    _inherit = 'billets.case'

    def make_slot_as_clone(self, cr, uid, ids, context=None):
        offer = self.browse(cr, uid, ids)[0]
        newArgs = {
            'event': offer.event.id,
            'male': offer.male,
            'female': offer.female,
            'notifystatus': 'none',
            'matchstatus': 'none',
            'cancelled': False,
            'days': [(6, 0, map(lambda x: x.id, offer.days))],
            'partner_id': offer.partner_id.id,
            'lacto_ovo_veg': offer.lacto_ovo_veg,
            'strict_vegan': offer.strict_vegan,
            'gluten_free': offer.gluten_free,
            'omnivore': offer.omnivore,
            'low_carb': offer.low_carb,
            'other_diets': offer.other_diets,
            'transport': offer.transport,
            'offer': offer.id
            }
        self.pool.get("billets.offered.slot").create(cr, uid, newArgs)

    def _check_matches(self, cr, uid, ids, field_name, arg, context):
        res = {}
        msd = dict(match_status)
        for offer in self.browse(cr, uid, ids):
            seen = {}
            for slot in offer.slots:
                seen[slot.matchstatus] = 1
            k = seen.keys()
            l = len(k)
            if l == 0:
                res[offer.id] = 'No slots'
            elif l == 1:
                res[offer.id] = 'All ' + msd[k[0]]
            else:
                res[offer.id] = map(lambda v: "Some %s " % msd[v], k)
        return res

    def _check_notifies(self, cr, uid, ids, field_name, arg, context):
        res = {}
        nsd = dict(notify_status)
        for offer in self.browse(cr, uid, ids):
            seen = {}
            for slot in offer.slots:
                seen[slot.notifystatus] = 1
            k = seen.keys()
            l = len(k)
            if l == 0:
                res[offer.id] = 'No slots'
            elif l == 1:
                res[offer.id] = 'All ' + nsd[k[0]]
            else:
                res[offer.id] = map(lambda v: "Some %s " % nsd[v], k)
        return res

    def _get_othercancelled(self, cr, uid, ids, field_name, arg, context):
        ret = {}
        for offer in self.browse(cr, uid, ids):
            anycancelled = False
            for slot in offer.slots:
                anycancelled = anycancelled or (slot.assignee and slot.assignee.cancelled)
            ret[offer.id] = anycancelled
        return ret

    _columns = {
          'male': fields.selection(flexibility_options, 'Male guest', size=16, required=True),
          'female': fields.selection(flexibility_options, 'Female guest', size=16, required=True),
          'slots': fields.one2many('billets.offered.slot', 'offer', 'Slots available'),
          'days': fields.many2many('billets.day', 'billets_offered_days', 'offer_id', 'day_id', 'Days'),
          'matchstatus': fields.function(_check_matches, fnct_search=brute_force_search, method=True, string='Match status', type='char'),
          'notifystatus': fields.function(_check_notifies, fnct_search=brute_force_search, method=True, string='Notify status', type='char'),
          'othercancelled': fields.function(_get_othercancelled, fnct_search=brute_force_search, method=True, string='Other cancelled', type='boolean')
        }

    _defaults = {
        'female': lambda *a: 'flex',
        'male': lambda *a: 'flex'
        }
billet_accommodation_offered()

class billet_accommodation_offered_slot(osv.osv):
    _name = 'billets.offered.slot'
    _description = 'Billet Accommodation Offered Slot'
    _inherit = 'billets.case'

    def _get_assignee(self, cr, uid, ids, field_name, arg, context):
        ret = {}
        for slot in self.browse(cr, uid, ids):
            wantids = self.pool.get('billets.wanted').search(cr, uid, [('provider', '=', slot.id)])
            ret[slot.id] = len(wantids) and wantids[0]
        return ret

    def _set_assignee(self, cr, uid, id, field_name, value, arg, context):
        bw = self.pool.get('billets.wanted')
        bw.write(cr, uid, [value], {'provider': id})

    _columns = {
          'male': fields.selection(flexibility_options, 'Male guest', size=16, required=True),
          'female': fields.selection(flexibility_options, 'Female guest', size=16, required=True),
          'notifystatus': fields.selection(notify_status, 'Notification status', size=25, required=True),
          'matchstatus': fields.selection(match_status, 'Match status', size=30, required=True),
          'cancelled': fields.boolean('Cancelled?', required=True),
          'days': fields.many2many('billets.day', 'billets_offered_slot_days', 'offered_slot_id', 'day_id', 'Days'),
          'offer': fields.many2one('billets.offered', 'Offer'),
          'assignee': fields.function(_get_assignee, None, _set_assignee, string='Assignee', method=True, type='many2one', relation='billets.wanted', fnct_search=brute_force_search)
        }
    _defaults = { 
        'female': lambda *a: 'flex',
        'male': lambda *a: 'flex',
        'notifystatus': lambda *a: 'none',
        'matchstatus': lambda *a: 'none',
        'cancelled': lambda *a: False }
billet_accommodation_offered_slot()

class billet_accommodation_wanted(osv.osv):
    _name = 'billets.wanted'
    _description = 'Billet Accommodation Wanted'
    _inherit = 'billets.case'

    def _get_othercancelled(self, cr, uid, ids, field_name, arg, context):
        ret = {}
        for wanted in self.browse(cr, uid, ids):
            ret[wanted.id] = (wanted.provider and wanted.provider.cancelled) == True
        return ret


    _columns = {
          'provider': fields.many2one(
            'billets.offered.slot',
            'Billet provider slot'),
          'notifystatus': fields.selection(notify_status, 'Notification status', size=25, required=True),
          'matchstatus': fields.selection(match_status, 'Match status', size=30, required=True),
          'gender': fields.selection([('male','Male'), ('female', 'Female'), ('other', 'Other')], 'Gender', size=10, required=True),
          'cancelled': fields.boolean('Cancelled?', required=True),
          'othercancelled': fields.function(_get_othercancelled, fnct_search=brute_force_search, method=True, string='Other cancelled', type='boolean'),
          'days': fields.many2many('billets.day', 'billets_wanted_days', 'wanted_id', 'day_id', 'Days')
        }
    _defaults = { 'notifystatus': lambda *a: 'none',
                  'matchstatus': lambda *a: 'none',
                  'cancelled': lambda *a: False }
billet_accommodation_wanted()
